/*
	*
	* This file is a part of libpapershell-wl.
	* Library for handling various services related to Paper Shell.
	* Copyright 2020 CuboCore Group
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <QDebug>

#include "DesQWindowManager.hpp"
#include "DesQWindowManagerImpl.hpp"

#include "wayland-wlr-foreign-toplevel-management-unstable-v1-client-protocol.h"

typedef struct ::zwlr_foreign_toplevel_manager_v1 wlrTopLevelManager;
typedef struct ::zwlr_foreign_toplevel_handle_v1 wlrTopLevelHandle;

DesQWindowManagerPrivate::DesQWindowManagerPrivate( wlrTopLevelManager *winMgr, struct wl_seat *seat, DesQWindowManager *ptr ) :
		QtWayland::zwlr_foreign_toplevel_manager_v1( winMgr ) {

	qDebug() << "DesQWindowManager constructor";

	q_ptr = ptr;
	m_seat = seat;

	m_tls.clear();
	m_tlHandleList.clear();
};

DesQWindowManagerPrivate::~DesQWindowManagerPrivate() {

	zwlr_foreign_toplevel_manager_v1_destroy( object() );
};

void DesQWindowManagerPrivate::zwlr_foreign_toplevel_manager_v1_toplevel( zwlr_foreign_toplevel_handle_v1 *handle ) {

	DesQWindowHandle *tlHandle = new DesQWindowHandle( handle );
	QObject::connect( tlHandle, &DesQWindowHandle::handleClosed, q_ptr, &DesQWindowManager::closeHandle );
	m_tlHandleList.append( tlHandle );

	Q_Q( DesQWindowManager );
	emit q->newToplevelHandle( tlHandle );
};

void DesQWindowManagerPrivate::zwlr_foreign_toplevel_manager_v1_finished() {

	// Does nothing
	qDebug() << "Toplevel Manager finished listening toplevel entries";
};

void DesQWindowManagerPrivate::closeHandle( DesQWindowHandle *handle ) {

	m_tlHandleList.removeAll( handle );

	// While desktop is showing manage the m_tls map
	if ( m_tls.count() )
		m_tls.remove( handle );
};

void DesQWindowManagerPrivate::storeAndMinimizeHandle( DesQWindowHandle *handle ) {

	m_tls[handle] = handle->appStates();

	if ( !( handle->appStates() & DesQWindowHandle::Minimized ) )
		handle->setMinimized();
};

DesQWindowManager::DesQWindowManager( zwlr_foreign_toplevel_manager_v1 *winMgr, wl_seat *seat, QObject *parent ) : QObject( parent ) {

	d_ptr = new DesQWindowManagerPrivate( winMgr, seat, this );
};

DesQWindowManager::~DesQWindowManager() {

	delete d_ptr;
};

DesQWindowHandleList DesQWindowManager::windowHandles() const {

	Q_D( const DesQWindowManager );
	return d->m_tlHandleList;
};

void DesQWindowManager::activateHandle( DesQWindowHandle *handle ) {

	Q_D( DesQWindowManager );
	handle->activate( d->m_seat );
};

void DesQWindowManager::closeHandle( DesQWindowHandle *tlHandle ) {

	Q_D( DesQWindowManager );
	d->closeHandle( tlHandle );
};

DesQWindowHandlePrivate::DesQWindowHandlePrivate( wlrTopLevelHandle *handle, DesQWindowHandle *ptr ) : QtWayland::zwlr_foreign_toplevel_handle_v1( handle ) {

	q_ptr = ptr;
	m_appID = "";
	m_appTitle = "";
	m_appStates = DesQWindowHandle::WindowStates();
};

DesQWindowHandlePrivate::~DesQWindowHandlePrivate() {

	zwlr_foreign_toplevel_handle_v1_destroy( object() );
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_title( const QString &title ) {

	m_appTitle = title;
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_app_id( const QString &app_id ) {

	m_appID = app_id;
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_output_enter( wl_output *output ) {

	Q_UNUSED( output )
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_output_leave( wl_output *output ) {

	Q_UNUSED( output )
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_state( wl_array *state ) {

	auto *states = static_cast<uint32_t *>( state->data );
	int numStates = static_cast<int>( state->size / sizeof( uint32_t ) );

	m_appStates = DesQWindowHandle::WindowStates();

	for ( int i = 0; i < numStates; i++ ) {
		if ( states[i] == ZWLR_FOREIGN_TOPLEVEL_HANDLE_V1_STATE_ACTIVATED ) {
			m_appStates |= DesQWindowHandle::Activated;
		} else if ( states[i] == ZWLR_FOREIGN_TOPLEVEL_HANDLE_V1_STATE_MAXIMIZED ) {
			m_appStates |= DesQWindowHandle::Maximized;
		} else if ( states[i] == ZWLR_FOREIGN_TOPLEVEL_HANDLE_V1_STATE_MINIMIZED ) {
			m_appStates |= DesQWindowHandle::Minimized;
		}
	}
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_done() {

	Q_Q( DesQWindowHandle );
	emit q->handleUpdated( q_ptr );
};

void DesQWindowHandlePrivate::zwlr_foreign_toplevel_handle_v1_closed() {

	Q_Q( DesQWindowHandle );
	emit q->handleClosed( q_ptr );
};

DesQWindowHandle::DesQWindowHandle( zwlr_foreign_toplevel_handle_v1 *handle, QObject *parent ) : QObject( parent ) {

	d_ptr = new DesQWindowHandlePrivate( handle, this );
};

DesQWindowHandle::~DesQWindowHandle() {

	delete d_ptr;
};

QString DesQWindowHandle::appID() const {

	Q_D( const DesQWindowHandle );
	return d->m_appID;
};

QString DesQWindowHandle::appTitle() const {

	Q_D( const DesQWindowHandle );
	return d->m_appTitle;
};

DesQWindowHandle::WindowStates DesQWindowHandle::appStates() const {

	Q_D( const DesQWindowHandle );
	return d->m_appStates;
};

void DesQWindowHandle::activate( wl_seat *seat ) {

	Q_D( DesQWindowHandle );
	d->activate( seat );
};

void DesQWindowHandle::setMaximized() {

	Q_D( DesQWindowHandle );
	d->set_maximized();
};

void DesQWindowHandle::unsetMaximized() {

	Q_D( DesQWindowHandle );
	d->unset_maximized();
};

void DesQWindowHandle::setMinimized() {

	Q_D( DesQWindowHandle );
	d->set_minimized();
};

void DesQWindowHandle::unsetMinimized() {

	Q_D( DesQWindowHandle );
	d->unset_minimized();
};

void DesQWindowHandle::close() {

	Q_D( DesQWindowHandle );
	d->close();
};
